// $Id: GoToActionHandler.java 16472 2012-10-23 16:41:41Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.*;
import javax.swing.*;
import java.awt.geom.*;
import java.awt.*;

/**
 * Create an action handler for "GoTo" actions and the named actions that move between
 * pages.
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">GoToActionHandler</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8
 */
public class GoToActionHandler extends ActionHandler {

    /**
     * Create a new GoToActionHandler
     * @since 2.11 just calls the public constructor
     */
    public GoToActionHandler() {
        super("GoToActionHandler");
    }

    public boolean matches(DocumentPanel panel, PDFAction action) {
        String type = action.getType();
        return type.equals("GoToFit") || type.equals("GoTo") || type.equals("GoToFitWidth") || type.equals("GoToFitHeight") || type.equals("GoToFitRectangle") || type.equals("Named:FirstPage") || type.equals("Named:PrevPage") || type.equals("Named:NextPage") || type.equals("Named:LastPage");
    }

    public void run(DocumentPanel panel, PDFAction action) {
        PDF pdf = panel.getPDF();
        int pagenumber = panel.getPageNumber();
        PDFPage page = null;
        float x = Float.NaN, y = Float.NaN, zoom = 0;
        final int PAD = 4;
        Rectangle2D fullsize = null;

        String type = action.getType();

        if (type.equals("GoTo")) {
            page = action.getPage();
            if (page!=null) {
                fullsize = PagePanel.getFullPageView(page);
                float[] coords = action.getGoToCoordinates();
                if (coords!=null && coords.length>=3) {
                    x = coords[0];
                    y = coords[1];
                    zoom = coords[2];
                }
            }
        } else if (type.equals("GoToFit")) {
            page = action.getPage();
            if (page!=null) {
                fullsize = PagePanel.getFullPageView(page);
                zoom = panel.getViewport().getTargetZoom(DocumentViewport.ZOOM_FIT, page);
            }
        } else if (type.equals("GoToFitWidth")) {
            page = action.getPage();
            if (page!=null) {
                fullsize = PagePanel.getFullPageView(page);
                zoom = panel.getViewport().getTargetZoom(DocumentViewport.ZOOM_FITWIDTH, page);
                float[] coords = action.getGoToCoordinates();
                if (coords!=null && coords.length>=1) {
                    y = coords[0];
                }
            }
        } else if (type.equals("GoToFitHeight")) {
            page = action.getPage();
            if (page!=null) {
                fullsize = PagePanel.getFullPageView(page);
                zoom = panel.getViewport().getTargetZoom(DocumentViewport.ZOOM_FITHEIGHT, page);
                float[] coords = action.getGoToCoordinates();
                if (coords!=null && coords.length>=1) {
                    x = coords[0];
                }
            }
        } else if (type.equals("GoToFitRectangle")) {
            page = action.getPage();
            if (page!=null) {
                fullsize = PagePanel.getFullPageView(page);
                float[] coords = action.getGoToCoordinates();
                if (coords!=null && coords.length>=4) {
                    Dimension avail = panel.getViewport().getViewportSize();
                    double availw = avail.getWidth() - PAD;
                    double availh = avail.getHeight() - PAD;
                    int dpi = Util.getScreenResolution(panel);
                    x = Math.min(coords[0], coords[2]);
                    y = Math.max(coords[1], coords[3]);
                    float w = Math.abs(coords[2]-coords[0]);
                    float h = Math.abs(coords[3]-coords[1]);
                    zoom = (float)Math.min(availw / w, availh / h) / dpi * 72;
                    // x and y are the top right of the rectangle
                }
            }
        } else if (type.equals("Named:NextPage")) {
            if (pagenumber < pdf.getNumberOfPages()-1) {
                page = pdf.getPage(pagenumber+1);
            }
        } else if (type.equals("Named:PrevPage")) {
            if (pagenumber > 0) {
                page = pdf.getPage(pagenumber-1);
            }
        } else if (type.equals("Named:FirstPage")) {
            page = pdf.getPage(0);
        } else if (type.equals("Named:LastPage")) {
            page = pdf.getPage(pdf.getNumberOfPages()-1);
        }
        if (page!=null) {
            if (pdf.getPages().contains(page)) {
                if (x==x) {
                    x -= fullsize.getMinX();
                    if (x<0 || x>fullsize.getWidth()) x = Float.NaN;
                }
                if (y==y) {
                    y = (float)(fullsize.getMaxY()-y);
                    if (y<0 || y>fullsize.getHeight()) y = Float.NaN;
                }
                if (zoom<0 || zoom>64) {
                    zoom = Float.NaN;
                }
                panel.setPage(page, x, y, zoom);
            } else {
                // Warning? Action refers to page from another document/deleted page
            }
        }
    }
}
