// $Id: ThumbnailExtractAction.java 17146 2013-03-20 14:47:37Z mike $

package org.faceless.pdf2.viewer2.feature;

import javax.swing.*;
import java.beans.*;
import javax.swing.event.*;
import java.awt.event.*;
import javax.swing.undo.*;
import java.text.MessageFormat;
import java.util.*;
import org.faceless.pdf2.*;
import org.faceless.pdf2.viewer2.*;

/**
 * This feature will allow pages to be extracted to a new Document via the {@link ThumbnailPanel}.
 * It is distinct from the drag/drop capability of the ThumbnailPanel, which allows pages to be
 * dragged between already open documents. This feature is only active when {@link MultiWindow} is
 * available, as it implies more than one document open in the Viewer.
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">ThumbnailExtractAction</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.12
 */
public class ThumbnailExtractAction extends ViewerFeature implements ThumbnailPanel.ThumbnailSelectionAction {

    public ThumbnailExtractAction() {
        super("ThumbnailExtract");
    }

    public Action getAction(final ThumbnailPanel.View view) {
        final DocumentPanel docpanel = view.getDocumentPanel();
        PDFViewer viewer = docpanel.getViewer();
        // We don't apply if the view isn't editable or we can't have multiple-windows
        if (!view.isFactoryEditable() || viewer == null || !viewer.hasFeature(MultiWindow.getInstance())) {
            return null;
        }

        final AbstractAction action = new AbstractAction(UIManager.getString("PDFViewer.Extract")) {
            public void actionPerformed(ActionEvent event) {
                final PDF pdf = docpanel.getPDF();

                if (JOptionPane.showConfirmDialog(docpanel, Util.getUIString("PDFViewer.ConfirmExtractPages", view.getSelectedPagesDescription()), view.getName(), JOptionPane.YES_NO_OPTION)==JOptionPane.YES_OPTION) {
                    JInternalFrame frame = (JInternalFrame)SwingUtilities.getAncestorOfClass(JInternalFrame.class, docpanel);
                    if (frame != null) {
                        List l = new ArrayList();
                        for (int i=0;i<view.getComponentCount();i++) {
                            ThumbnailPanel.SinglePagePanel panel = (ThumbnailPanel.SinglePagePanel)view.getComponent(i);
                            if (panel.isSelected()) {
                                l.add(panel.getPage());
                                panel.setSelected(false);
                            }
                        }
                        pdf.getPages().removeAll(l);
                        PDF newpdf = new PDF();
                        newpdf.getPages().addAll(l);
                        docpanel.setPage(pdf.getPage(0));
                        String loadTitle = UIManager.getString("PDFViewer.ThumbnailExtractAction.loadTitle");
                        loadTitle = MessageFormat.format(loadTitle, null, frame.getTitle());
                        docpanel.getViewer().loadPDF(newpdf, loadTitle);
                        // Not strictly correct, this will just close the other document - it won't
                        // put the pages back.
                        docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, Undo.DISCARD));
                    }
                }
            }
        };

        view.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent event) {
                if (event.getPropertyName().equals("selection") || event.getPropertyName().equals("selected")) {
                    // We're only enabled if the action won't create or leave an empty document
                    PDF pdf = view.getDocumentPanel().getPDF();
                    int select = 0;
                    int total = pdf == null ? 0 : pdf.getNumberOfPages();
                    for (int i=0;i<view.getComponentCount();i++) {
                        ThumbnailPanel.SinglePagePanel panel = (ThumbnailPanel.SinglePagePanel)view.getComponent(i);
                        if (panel.isSelected()) {
                            select++;
                        }
                    }
                    action.setEnabled(view.isEditable() && select > 0 && select < total);
                } else if (event.getPropertyName().equals("ancestor") && event.getNewValue() == null) {
                    view.removePropertyChangeListener(this);
                }
            }
        });

        view.getDocumentPanel().addDocumentPanelListener(new DocumentPanelListener() {
            public void documentUpdated(DocumentPanelEvent event) {
                if ("permissionChanged".equals(event.getType())) {
                    action.setEnabled(action.isEnabled() && view.isEditable() && view.isDraggable());
                }
            }
        });

        return action;
    }

}
